//++
//player.c - system startup, POST, and initialization
//
// Copyright (C) 2005 by Spare Time Gizmos.  All rights reserved.
//
// This file is part of the Spare Time Gizmos' MP3 Player firmware.
//
// This firmware is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//
//DESCRIPTION:
//
//TODO:
//
// REVISION HISTORY:
// dd-mmm-yy    who     description
// 15-May-05    RLA     New file.
// 12-Oct-05	RLA	SDCC doesn't support printf precision for strings
//			The SDCC __DATE__ macro returns too many characters
//			On non-FAT volumes, don't display the volume label
//			Change calling sequences for FAT and CFCARD routines
// 15-Oct-05	RLA	Add VOLUME conditional
// 19-Oct-05	RLA	Call ResetMP3Decoder() to abort playing (when the knob
//			 is turned).  This avoids noise in the output.
//--

// Include files...
#include <stdio.h>		// needed so DBGOUT(()) can find printf!
#include <string.h>		// needed for memcpy(), memset(), et al...
#include "reg66x.h"		// declarations for Philips 89C66x processors
#include "standard.h"		// standard types - BYTE, WORD, BOOL, etc
#include "player.h"		// project wide (hardware configuration) declarations
#include "post.h"		// power on self test diagnostics
#include "debug.h"		// debuging (serial port output) routines
#include "timer.h"		// time keeping functions
#include "buffer.h"		// disk buffer management functions
#include "display.h"		// LCD/VFD display interface
#include "i2c.h"		// I2C driver functions
#include "sta013.h"		// STA013 MP3 decoder functions
#include "cfcard.h"		// compact flash card interface functions
#include "fat.h"		// FAT12/16 functions
#include "switches.h"		// user interface (rotary encoder) functions
#include "id3.h"		// ID3 tag definitions and functions

//   This is the copyright notice, version, and date for the software in plain
// ASCII.  Even though this only gets printed out in the debug version, it's
// always included to identify the ROM's contents...
PUBLIC char const code g_szFirmware[] =
  "MP3 Player Firmware " __DATE__ " " __TIME__;
PUBLIC char const code g_szCopyright[] =
  "Copyright (C) 2005 by Spare Time Gizmos.    All rights reserved.";



//++
//   This routine will show the copyright "screen" on the display.  It will
// pause for five seconds before returning...
//--
PRIVATE void ShowCopyright (void)
{
  ClearDisplay();
  DisplayTextAt(0, 0, COPYRIGHT_SYMBOL " SPARE TIME GIZMOS");
  DisplayTextAt(1, 0, "MP3 Player v%03u %04X", (WORD) VERSION, g_wROMChecksum);
  DisplayTextAt(2, 0, "   %1bdK RAM %2bdK ROM",
   (BYTE) ((RAMSize()+256) >> 10), (BYTE) (ROMSIZE >> 10));
  DisplayTextAt(3, 4, "%s", __DATE__);
  DelayMS(5000);
  //while (PUSH_BUTTON) ;
}

//++
//   This routine will clear the display and show the first three (of the four)
// lines that appear while we're playing an MP3.  These lines show the file name
// and (if the file has an ID3 tag) the artist and song name.  The remaining
// fourth line comes from ShowPlaybackStatus()...
//--
PRIVATE void ShowPlaybackNames (PXDIRENT pxDirEnt)
{
  ClearDisplay();
  //   If this file has an LFN (long file name) then show that on the first
  // line; otherwise the traditional 8.3 name will have to do.  Note that
  // we never show the .MP3 extension in any case - when all files are MP3,
  // why bother?
  if (strlen(g_szLongFileName) > 0) {
    DisplayTextAt(0,0,"%.20s", g_szLongFileName);
  } else {
    DisplayTextAt(0,0,"%8.8s.%3.3s", pxDirEnt->szFileName, pxDirEnt->szFileType);
  }
  //   Now open the file and attempt to read the ID3 tag.  If we get lucky and
  // find one, then use that to show the title and artist.
  if (ReadID3Tag(pxDirEnt)) {
    DisplayTextAt(1,0,"%.20s",g_szArtistName);
    DisplayTextAt(2,0,"%.20s",g_szTrackName);
  }
}

//++
//   And this routine will display the fourth line used while we're playing.
// This line shows the current running time (in mmm:ss), the MP3 bit rate
// (kbps) and the sampling frequency (kHz)...
//--
PRIVATE void ShowPlaybackStatus (BOOL fPaused)
{
  BYTE bSeconds;  WORD wMinutes;
  GetElapsedTime(wMinutes);
  bSeconds = (BYTE) (wMinutes % 60);  wMinutes /= 60;
  if (fPaused)
    DisplayTextAt(3,0,"%3u:%02bu    PAUSED    ", wMinutes, (BYTE) bSeconds);
  else
    DisplayTextAt(3,0,"%3u:%02bu %3ukbps %2bukHz", 
      wMinutes, (BYTE) bSeconds, GetMP3BitRate(), (BYTE) GetMP3SampleFrequency());
}


//++
//   Unfortunately, we don't support hot swapping of CF cards (the CF standard
// makes that pretty much impossible for any application that uses TrueIDE mode)
// so contrary to what its name might lead you to believe, this routine does NOT
// wait for a card to be inserted.  If one isn't already in there at power up,
// this routine just displays an error message and returns FALSE, and that will
// make the main program loop forever until the user reboots us.
//
//   If a CompactFlash card is inserted now, then this routine will initialize
// the IDE interface and attempt to find a FAT16 filesystem on it.  If we're
// successful it will display a brief status screen (for about 2 seconds) and
// then return TRUE; if the card can't be read or if it doesn't contain a FAT16
// file system, then an error message is displayed and FALSE results.
//--
PRIVATE BOOL IsCardReady (void)
{
  ClearDisplay();

  //   If there's no CF card in the socket now, then tell him to turn off the
  // power, insert one, and try again...
  if (!IsCardInserted()) {
    DBGOUT(("??IsCardReady: No card inserted ...\n"));
    //                   01234567890123456789
    DisplayStringAt(0,0,"???? NO CF CARD ????");
  //DisplayStringAt(1,0,"                    ");
    DisplayStringAt(2,0,"  insert card and   ");
    DisplayStringAt(3,0,"    cycle power     ");
    return FALSE;
  }

  //   Initialize the IDE interface.  If we can't, then either the card is bad
  // or he has tried to hot swap it (which prevents the card from going into
  // TrueIDE mode).  Either way he has to restart...
  if (!InitializeCard()) {
    DBGOUT(("??IsCardReady: Failed to initialize TrueIDE interface ...\n"));
    //                   01234567890123456789
    DisplayStringAt(0,0,"????? BAD CARD ?????");
  //DisplayStringAt(1,0,"                    ");
    DisplayStringAt(2,0,"unable to initialize");
    DisplayStringAt(3,0," TrueIDE interface  ");
    return FALSE;
  }

  //  Identify the card's manufacturer, type and size.  Mostly this is just for
  // show, but it does serve to verify that the card really is a Flash Memory
  // device (after all, it might be a modem or a network card or anything!)...
  if (!IdentifyCard()) {
    DBGOUT(("??IsCardReady: Unrecognized card \"%s\"\n", g_szCardName));
    //                   01234567890123456789
    DisplayStringAt(0,0,"????? BAD CARD ?????");
  //DisplayStringAt(1,0,"                    ");
    DisplayStringAt(2,0," unable to identify ");
    DisplayStringAt(3,0,"    card  type      ");
    return FALSE;
  }
  DisplayTextAt(0, 0, "%.20s", g_szCardName);
  DisplayTextAt(1, 0, "Type %04X Size %uMb", g_wCardAttributes, g_wCardSize);

  // Finally, try to find a file system that we can deal with...
  if (!InitializeFAT()) {
    DBGOUT(("??IsCardReady: Not a FAT16 file system \"%s\"\n", g_szSystemType));
    //                   01234567890123456789
    DisplayStringAt(2,0,"??? not a FAT16  ???");
    DisplayStringAt(3,0,"??? file system  ???");
    //DisplayTextAt(1, 10, "%s", g_szSystemType);
    return FALSE;
  }

  //   It's a good card and a good file system.  Find the volume label and
  // count the number of MP3 files in the root directory.  If there are no
  // MP3 files, then there's nothing we can do with this card!
  GetVolumeData();
  DisplayTextAt(2,0,"%11.11s", g_szVolumeLabel);
  if (g_wTotalMP3Files == 0) {
    //                   01234567890123456789
    DisplayStringAt(3,0,"??? no mp3 files ???");
    return FALSE;
  }    
  DisplayTextAt(3,0,"%u MP3 Files", g_wTotalMP3Files);

  // Delay a few moments so the user can read the screen and then return...
  DelayMS(5000);
  return TRUE;
}

//++
//   This routine will play one MP3 file given its directory entry.  Ordinarily
// it will play the entire file, beginning to end, and won't return until the
// end of file is reached.  A basic pause/resume function, using the push button,
// is implemented locally.  If, however, the rotary encoder is turned this
// routine will abort playback of the current file and return immediately.  It
// doesn't change the value of g_nEncoderCount in the process of doing this -
// it's up to the caller to notice that this value is non-zero and skip forward
// or backwards as required.
//--
PUBLIC void PlayOneMP3 (PXDIRENT pxDirEnt)
{
  PIBCB pBCB;  BOOL fFinished;  BOOL fPaused;
#ifdef DEBUG
  BYTE bFreeBufferCount;
#endif

  //   This is handy for debugging memory leaks in the player, not that anything
  // like that has ever happened to _me_ !
#ifdef DEBUG
  bFreeBufferCount = ShowBufferPool();
  //DBGOUT(("PlayOneMP3: Starting with %bd buffers\n", bFreeBufferCount));
#endif

  //   Open the MP3 file, initialize the STA013 data pump, start the STA013,
  // and then start the elapsed time counter.  Since the STA013 doesn't have
  // any feature to actually report the current running time of the song, we
  // simply time it independently of the STA013...
  DBGOUT(("PlayOneMP3: playing file %8.8s.%3.3s ...\n", pxDirEnt->szFileName, pxDirEnt->szFileType));
  OpenFile(pxDirEnt);
  InitializeMP3DRQ();  fPaused = FALSE;
  StartMP3Decoder();
  ResetElapsedTime();  StartElapsedTimer();

  //   Keep playing until (what else??) we're finished.  The main purpose of
  // this loop is to keep the STA013 supplied with data from the MP3 file.
  // The secondary purpose, which we do as time permits, is to "pump" the
  // UI - update the LCD/VFD display and check the knob/button for activity.
  for (fFinished = FALSE;  !fFinished;  ) {

    //   Finally, here it is, all the magic happens here.  The basic plan is
    // that the background code (that's us!) continuously trys to remove empty
    // disk buffers from the free list, read the next file sector into them,
    // and then put the filled buffers on the STA013 play queue.  The foreground
    // task (that's the MP3DRQ interrupt routine) removes filled disk buffers
    // from the play list and sends their data to the STA013.  So long as we
    // stay ahead of them, everything will be fine...
    INT_OFF;  ALLOCATE_BUFFER(pBCB);  INT_ON;
    if (pBCB != NULL) {
      pBCB->cbBuffer = ReadFile(pBCB->pbBuffer);
      fFinished = (pBCB->cbBuffer) < IDE_SECTOR_SIZE;
      //DBGOUT(("PlayOneMP3: read buffer pBCB=0x%02bX, pbBuffer=0x%04X, cbBuffer=%u\n",
      //  (BYTE) pBCB, (WORD) pBCB->pbBuffer, pBCB->cbBuffer));
      EX0 = 0;  ADD_TO_QUEUE(pBCB,g_pPlayListHead,g_pPlayListEnd);  EX0 = 1;
    }

    //   If the elapsed time timer has ticked (it happens once a second), then
    // update the display with the current runtime, bit rate, and other status.
    // Note that in the case of VBR files, the bit rate isn't constant and will
    // change with every update!
    if (IsTimerChanged()) ShowPlaybackStatus(fPaused);

    //   In the current UI, the button is a simple pause/resume function (push
    // it once and the playback pauses; push it again and playback resumes).
    // This operation is implemented entirely here - when the button is pushed
    // we pause the STA013 but continue looping.  Eventually all the free buffers
    // will be filled and placed on the play list and nothing more will happen
    // until the button is pressed again.  When that happens the STA013 is
    // restarted and playback continues...
    if (g_bButtonTime > 0) {
      //   Let's think about this a minute - the button is down right this moment.
      // If we simply process that fact and do nothing more, then we'll come right
      // back here and think that the button was pressed a second time.  To prevent
      // this we have to wait for the button to be released, but waiting prevents
      // any disk buffers from being filled!   Turns out that this doesn't matter
      // much - if we're playing now, then the STA013 will go silent when it runs
      // out of data, but we're going to pause anyway so nobody will notice.  If
      // we paused now, then the STA013 isn't playing anyway and it doesn't need
      // any data...
      while (g_bButtonTime > 0) ;
      // Now that that's over, and the rest is easy...
      fPaused = !fPaused;
      if (fPaused) {
        PauseMP3();  StopElapsedTimer();
	DBGOUT(("PlayOneMP3: playback paused ...\n"));
      } else {
        ResumeMP3();  StartElapsedTimer();
	DBGOUT(("PlayOneMP3: playback resumed ...\n"));
      }
      // Update the display to show PAUSED or the current status...
      ShowPlaybackStatus(fPaused);
    }

    //   If the encoder has been turned, then the user wants to skip to either
    // the next or previous song.  Doesn't matter to us - we simply abort the
    // current song and let the caller worry about that...
    if (g_nEncoderCount != 0) {
      ResetMP3Decoder();  FlushPlayList();  fFinished = TRUE;
      DBGOUT(("PlayOneMP3: playback aborted ...\n"));
    }
  }

  // WAIT for the play queue to drain!!
  while (!IsPlayListEmpty()) ;
  StopMP3Decoder();
  ASSERT( (ShowBufferPool() == bFreeBufferCount), ("PlayOneMP3: BUFFERS HAVE DISAPPEARED!!\n") );
  CloseFile();
  DBGOUT(("PlayOneMP3: finished playing %8.8s.%3.3s\n", pxDirEnt->szFileName, pxDirEnt->szFileType));
}


PUBLIC void PlayAllMP3s (void)
{
  PXDIRENT pxDirEnt;
  pxDirEnt = FirstMP3File();
  ShowPlaybackNames(pxDirEnt);
  while (pxDirEnt != NULL) {
    PlayOneMP3(pxDirEnt);
    if (g_nEncoderCount > 0) {
      while (g_nEncoderCount > 0) {
        pxDirEnt = NextMP3File();
	if (pxDirEnt == NULL) return;
        ShowPlaybackNames(pxDirEnt);
        if (--g_nEncoderCount == 0) DelayMS(500);
      }
    } else if (g_nEncoderCount < 0) {
      ShowPlaybackNames(pxDirEnt);
      if (++g_nEncoderCount == 0) DelayMS(750);
      while (g_nEncoderCount < 0) {
        pxDirEnt = PrevMP3File();
	if (pxDirEnt == NULL) return;
        ShowPlaybackNames(pxDirEnt);
        if (++g_nEncoderCount == 0) DelayMS(500);
      }
    } else {
      pxDirEnt = NextMP3File();
      ShowPlaybackNames(pxDirEnt);
    }
  }
}


//++
//   This routine initializes the system and all modules, both hardware and
// software.  It's only ever called once, as the first thing in main().  Note
// that the POST is implict in many of these routines - for examine, Initialize-
// STA013() also tests the STA013 and POSTs an error code if it's not working.
//--
PRIVATE void InitializeSystem (void)
{
  P0 = P1 = P2 = P3 = 0xFF;
  //   The LED is carefully arranged so that it is ON after a CPU reset (that's
  // because RESET floats all the I/O port pins, and a HIGH output turns the
  // LED on).  By turning it OFF as the first thing, it lets us know that the
  // CPU is at least alive and running...
  LED_OFF;

  //   If this is the debug version then turn on the 89C420's secondary serial
  // port and say hello...
#ifdef DEBUG
  InitializeDebugSerial();
  printf("\n\n%s V%03u ROM %bdK Checksum %04X\n%s\n\n",
    g_szFirmware, (WORD) VERSION, (BYTE) (ROMSIZE >> 10), g_wROMChecksum, g_szCopyright);
#endif

  // Test the CPU's internal flash and verify the code checksum...
  if (ROMChecksum(ROMSIZE) != g_wROMChecksum) POST(PC_CHECKSUM, FALSE);
  DBGOUT(("Flash ROM checksum OK ...\n"));

  //  Turn off ALE and enable the 89C66x internal XRAM.  Test the size of the
  // SRAM and if there aren't at least 1792 bytes of internal RAM (a P89C664)
  // then just POST and give up.  There's not enough room for disk buffers in
  // anything less...
  AUXR = AO;
  DBGOUT(("XRAM size %u bytes ...\n", RAMSize()));
  if (RAMSize() < 1792) POST(PC_SRAM, FALSE);

  //  Initialize all interrupts to low priority, disable all the individual
  // interrupt sources, and then set the global interrupt enable.  This allows
  // individual devices to turn on their own interrupt enable bits as soon as
  // they are ready...
  IP = IPH = IE = IE1 = 0;  EA = 1;

  // Initialize the other software and hardware modules....
  InitializeTimer();   		// the interval timer and elapsed time meter
  InitializeDisplay();		// the LCD or VF display
  InitializeSwitches();		// the rotary encoder and push button polling
  InitializeSTA013();  		// the (in)famous STA013

  //   This call will initialize the free buffer pool in XDATA RAM and the parameter
  // specifies how many bytes of XDATA to reserve for static variables (i.e. anything
  // you may see in one of these modules declared with XDATA!).  This value is really
  // just an approximation - it doesn't matter so long as it's at least as big as
  // what we actually use.  It'd be nice if there was a way to have the linker set
  // this for us, but I don't know how.
  //
  //   By the way, there's no point in trying to fine tune this value - since
  // buffers are always 512 bytes long, unless you can reduce it to 256 bytes 
  // or less (pretty much impossible with the current code), there won't be
  // enough room for another buffer anyway.  In fact, any value between 256 and
  // 767 will give the same results!
  InitializeBuffers(512+256-1);

  // That's it - we're as ready as we'll ever be...
  DBGOUT(("System initialization complete ...\n\n"));
}


//++
// System initialization and startup...
//--
void main (void)
{
  InitializeSystem();
  ShowCopyright();

  //   Make sure a readable CF card is inserted.  Since there's currently no way
  // to hot swap cards, if there's anything wrong with the one we've got all we
  // can do is loop forever until the power is turned off.
  if (!IsCardReady()) HALT;
  DBGOUT(("Playing (%u MP3 files found) ...\n\n", g_wTotalMP3Files));

  while (TRUE) PlayAllMP3s();
}

